
/*
 * demonstrate using the timer '2' of the 8254 of the motherboard.
 */

#ifdef __USAGE
%C 
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <i86.h>
#include <conio.h>
#include <sys/timers.h>
#include <sys/sched.h>
#include <sys/kernel.h>

typedef	unsigned short SAMPLE;

/* timer routines */
void fast_timer_init( void );
SAMPLE fast_timer( void );
void fast_timer_calibrate( void );
int fast_timer_overflow( void );
long fast_timer_max( void );
void fast_timer_restore( void );

int			verbose = 2;	/* Flag controlling verbose output			*/
/* these values are in seconds */
SAMPLE		ftic_overhead;	/* Tic count to call fast_timer() routine	*/
double		ftic;			/* Nanosecond interval of fast tick			*/
double		ftic_max;		/* Maximum measurable interval				*/
double		ftic_usec;		/* Microseconds per fast_timer() call		*/

/*
Disable interrupts, but return the flag word so that the previous int enable
status can be properly restored.
*/
unsigned disable(void);
#pragma aux disable = "pushf" "cli" "pop ax" parm [] modify [ax];

/*
Enable interrupts, but return the flag word so that the previous int enable
status can be properly restored.
*/
unsigned enable(void);
#pragma aux enable = "pushf" "sti" "pop ax" parm [] modify [ax];

/*
Restore interrupt enable status to what it was before disable/enable was called.
*/
void restore(unsigned flags);
#pragma aux restore = "push ax" "popf"	parm [ax] modify [];

#define USECS_PER_SEC 1000000

void main(argc,argv)
int argc;
char *argv[];
{
int i;
int c;
SAMPLE data[1000];
SAMPLE start, stop;
unsigned int samples = 10;

while ( ( c = getopt( argc, argv, "s:" ) ) != -1 ) {
      switch(c) {
         case 's' : samples = atoi(optarg);
                    break;
         }
      }

if ( setprio( getpid(), 29 ) == -1 ) {
   fprintf(stderr,"Error setting priority.\n");
   exit(-1);
   }

atexit( fast_timer_restore );
fast_timer_init();
fast_timer_calibrate();

printf("tic count to call fast_timer     = %u\n",
        ftic_overhead);
printf("nanosecond interval of fast tick = %16.8f usecs.\n",
        ftic * USECS_PER_SEC);
printf("timing a call to Yield.\n");

for ( i=0; i < samples; i++ ) {
    fast_timer_init();
    start = fast_timer();
    Yield();
    stop  = fast_timer();
    data[i] = (start - stop);
    if ( fast_timer_overflow() ) {
       printf("Danger!! overflow of fast timer! Values not correct.\n");	
       }
    }
printf("fast ticks : usec time\n");
for ( i=0; i < samples; i++ ) {
    printf("%u :%15.8f usecs  adjust %u :%15.8f usec\n",
            data[i], 
            data[i] * ftic * USECS_PER_SEC,
            data[i] - ftic_overhead,
            (data[i]-ftic_overhead) * ftic * USECS_PER_SEC
            );
    }
}

/***************************************************************************/
/*
This uses the timer '2' of the 8254 on the motherboard. The *assumption* is 
that this timer is not being modified by any system software. If this 
assumption does not hold, then this will generate sporadic results and
perhaps confuse the OS timing.
*/

#define _8254BASE		0x40
#define _8254COUNT(x)	((x) + _8254BASE)
#define _8254CONTROL	(_8254BASE + 3)
#define _8254MODE(x)	(((x)&0x07) << 1)
#define _8254RW(x)		(((x)&0x03) << 4)
#define _8254LATCH		(0)
#define _8254SELECT(x)	(((x)&0x03) << 6)
#define _GATECH3		(0x61)
#define	_GATECH3_VAL	(0x01)

#define RELOADVALUE     0xffff

/*
Reset fast timer to be able to measure full interval.
*/
void
fast_timer_init() {
	unsigned x;

	x = disable();
	outp(_GATECH3,_GATECH3_VAL);
	/* set counter 2 to mode 3, read/write both */
	outp(_8254CONTROL, _8254SELECT(2) | _8254MODE(0) | _8254RW(3));
	outp(_8254COUNT(2), (unsigned short)RELOADVALUE >> 8);
	outp(_8254COUNT(2), (unsigned short)RELOADVALUE & 0xff);
	restore( x );
	}

/*
Return the current value from the fast timer.  Be careful, this routine is
called by the interrupt handler as well.  The disable/restore pragmas are
used so that this code can be called at process or interrupt time.
*/
SAMPLE
fast_timer() {
	unsigned y;
	SAMPLE x;

	y = disable();
	outp(_8254CONTROL, _8254SELECT(2) | _8254LATCH);
	x = inp(_8254COUNT(2));
	x |= inp(_8254COUNT(2)) << 8;
	restore( y );
	return( x );
	}

/*
Return a true status if the interval timer overflowed.
*/
int
fast_timer_overflow() {
	outp( _8254CONTROL, 0xe8 );
	return( (inp( _8254COUNT(2) ) & 0x80) != 0 );
	}

/*
Return the maximum number of fast tics the counter can count.
*/
long
fast_timer_max() {
	return( RELOADVALUE );
	}

/*
Set fast timer back to what it was before rtbench ran.  This allows the
beep to operate normally again.
*/
void
fast_timer_restore() {
	/* @@@ timer restore code	*/
	}

/****************************************************************************/
/*
 * calibration routines
 */

/*
Return time in microseconds that it took for 'iter' calls to fast_timer()
*/
long
fast_tick_measure( long iter ) {
	long i;
	struct timespec start, stop;
/*
	We need a volatile variable here so that the timing loop will always do
	the assignment (and write a value to memory, not a register).
	@@@ Unroll the loop
*/
	static volatile SAMPLE t = 0;

	getclock( TIMEOFDAY, &start );
	for( i = 0; i*10 < iter; ++i ) {
		t = fast_timer();		/* Unroll the loop for better accuracy */
		t = fast_timer();
		t = fast_timer();
		t = fast_timer();
		t = fast_timer();
		t = fast_timer();
		t = fast_timer();
		t = fast_timer();
		t = fast_timer();
		t = fast_timer();
		}
	getclock( TIMEOFDAY, &stop );
	return( ((stop.tv_sec - start.tv_sec) * 1000000L) +
			(stop.tv_nsec - start.tv_nsec)/1000L );
	}

void
fast_timer_calibrate() {
	long iter, ft;

	if (verbose)
		printf( "------Calibrating interval timer----------------\n" );
	iter = 200000L;
	/* Keep increasing 'iter' until at least a million microseconds elapse	*/
	while ( (ft = fast_tick_measure( iter )) < 1000000L ) iter *= 2L;
	ftic_usec = (double) ft / (double) iter;

	/* Compute how many fast ticks it takes to read the fast tick timer		*/
	/* Since the PC crystal is 1.19319 MHz, this can be accurately known	*/
	ftic_overhead = (SAMPLE) ((ftic_usec * 1193180L)/1000000L);

	/* Compute the interval per fast_tick as 1/frequency of the crystal		*/
	ftic = (double) 1.0/1.19318E6;
	/* Compute the maximum measurable interval with this fast timer			*/
	ftic_max = ftic * fast_timer_max();

	if (verbose) {
		printf( "Fast Timer (ftick) Characteristics:\n" );
		printf( "Precision          = %f nanoseconds/ftick\n",
			ftic * 1E9 );
		printf( "Call overhead      = %f microseconds (%u fticks)\n",
			ftic_usec, (unsigned)ftic_overhead );
		printf( "Maximum interval   = %f microseconds\n",
			ftic_max * 1E6 );
		printf( "------Calibration complete----------------------\n\n\n" );
		}
	}

